// -----------------------------------------------
// KBall Game
// By Kronoman
// Copyright (c) 2003-2004, Kronoman
// In loving memory of my father
// -----------------------------------------------
// gkernel.h
// -----------------------------------------------
// Game main loop and kernel
// ----------------------------------------------- 

#include "gkernel.h" // kernel header

#include "gerror.h" // to show any error that may arise 
#include "misc_def.h" // some definitions
#include "filehelp.h"
#include "savescrs.h" // for screenshoots

// ----------------------------------------------- 
// Global variables
// ----------------------------------------------- 
// Counter for the main timer
static volatile int speed_counter = 0;

// this 2 counters are for count the fps
static volatile int frame_count, fps;

// ---- End of global variables declaration ---- 

// ----------------------------------------------- 
// Timing stuff
// ----------------------------------------------- 
// Timer callback for the speed counter
void increment_speed_counter()
{
   speed_counter++;
}
END_OF_FUNCTION(increment_speed_counter);

// Timer callback for measuring the frames per second
void fps_proc(void)
{
   fps = frame_count;
   frame_count = 0;
}                                                                               
END_OF_FUNCTION(fps_proc);

// ==========================================================================
// KERNEL CLASS ITSELF BELOW THIS
// ==========================================================================

int CGameKernel::timer_installed_count = 0; // how many times we installed the timer; when this is 0, remove/install timers

// -------------------------------------
// Constructor
// -------------------------------------
CGameKernel::CGameKernel()
{
	mtracer.add("CGameKernel::CGameKernel()");
	
	game_over = false;
	
	dbuffer = NULL; // doble buffer
	
	backdropbmp = NULL; // background bitmap for current level
	
	current_level_file_name[0] = '\0';
}

// -------------------------------------
// Destructor
// -------------------------------------
CGameKernel::~CGameKernel()
{
	mtracer.add("CGameKernel::~CGameKernel()");

	this->shutdown();
}

// -------------------------------------------------------------------
// Initializes game (sets timers, loads data, etc)
// -------------------------------------------------------------------
void CGameKernel::init()
{
	mtracer.add("CGameKernel::init()");

	if (dbuffer == NULL) dbuffer = create_bitmap(SCREEN_W, SCREEN_H); // create doble buffer bitmap
	
	if (dbuffer == NULL) raise_error("CGameKernel::init()\nERROR: unable to create doble buffer bitmap!\n");

	clear_bitmap(screen);
	textout(screen, font, "Please wait... loading data...", 0, 0, makecol(255,255,255));
	
	// Load general data for the game
		char tmp_file_buf[2048];
		
		if (game_map.load_tile_set_from_file(where_is_the_filename(tmp_file_buf,TILESET_FILE)) ) raise_error("CGameKernel::init()\nERROR!\nCan't load  \"%s\" file!\n", TILESET_FILE); // load tileset
		
		if (main_data_file.load_datafile(where_is_the_filename(tmp_file_buf,SPRITES_FILE)) ) raise_error("CGameKernel::init()\nERROR!\nUnable to load \"%s\" file!\n", SPRITES_FILE);

	mtracer.add("\tGeneral data loaded OK");
		
	// find the ball's sprite texture and mask
	player_ball.spr = (BITMAP *)main_data_file.get_resource_dat("BALL");
	if (player_ball.spr == NULL) raise_error("CGameKernel::init()\nERROR! Can't find BALL sprite!\n");
	
	player_ball.spr_shadow = (BITMAP *)main_data_file.get_resource_dat("BALL_MASK_SHADOW");
	if (player_ball.spr_shadow == NULL) raise_error("CGameKernel::init()\nERROR! Can't find BALL_MASK_SHADOW sprite!\n");
	
	mtracer.add("\tBall data loaded OK");

	// find GUI fonts
	game_time_font = (FONT *)main_data_file.get_resource_dat("GAME_TIME_FONT");
	game_score_font = (FONT *)main_data_file.get_resource_dat("GAME_SCORE_FONT");
	game_messages_font = (FONT *)main_data_file.get_resource_dat("GAME_MESSAGES_FONT");
	
	// set some stuff
	game_map.timer_tick_rate = GKERNEL_FPSQ; // set timer ratio for map

	// set the timers
	if (timer_installed_count == 0)
	{
		LOCK_VARIABLE(speed_counter);
		LOCK_FUNCTION((void *)increment_speed_counter);   
	
		LOCK_VARIABLE(fps);
		LOCK_VARIABLE(frame_count);
		LOCK_FUNCTION((void *)fps_proc);
	
		// install the timer
		if (install_int_ex(increment_speed_counter, BPS_TO_TIMER(GKERNEL_FPSQ))) raise_error("CGameKernel::init() : Can't install timer at %d bps\n", GKERNEL_FPSQ);
		if (install_int(fps_proc, 1000)) raise_error("CGameKernel::init() : Can't install timer to measure fps.");
		
		mtracer.add("\tTimer installed for first time");
	}
	
	timer_installed_count++; // just to make sure that nobody else will remove our precious timer until we are ready too
	
	mtracer.add("\ttimer_installed_count = %d", timer_installed_count);
	
	// here start all the variables and all stuff
	game_over = false;
	frame_count = fps = 0;
	
	current_level_file_name[0] = '\0';
}

// -------------------------------------------------------------------
// shuts down game (unsets timers, unloads data, etc)
// -------------------------------------------------------------------
void CGameKernel::shutdown()
{
	mtracer.add("CGameKernel::shutdown()");
	
	// remove timer(s), only if active
	if (timer_installed_count > 0)
	{
		timer_installed_count--;
		
		mtracer.add("\ttimer_installed_count = %d", timer_installed_count);
		
		if (timer_installed_count == 0)
		{
			remove_int(increment_speed_counter);
			remove_int(fps_proc);
			timer_installed_count = 0;
			
			mtracer.add("\tTimer totally removed");
		}
	}
	
	// FREE THE RAM USED
	game_map.free_memory();

	main_data_file.nuke_datafile();

	particle_manager.nuke_particle_list();

	if (dbuffer != NULL)
	{
		destroy_bitmap(dbuffer);
		dbuffer = NULL;
	}
	
	if (backdropbmp != NULL)
	{
		destroy_bitmap(backdropbmp);
		backdropbmp = NULL;
	}
	
	current_level_file_name[0] = '\0';
}

// -------------------------------------------------------------------
// This is the main game loop
// 
// **NOTICE**
// YOU **MUST** CALL IN THIS ORDER:
//   gkernel.init();
//   gkernel.load_level_file("level.map");
//   gkernel.game_loop();
//   gkernel.shutdown();
//
// will return CBALL_IS_DEAD, or CBALL_EXIT_LEVEL (dead, or won level)
//
// **NOTE** THE SPECIAL RETURN VALUE 'GKERNEL_USER_FINISHED_GAME' MEANS THAT THE PLAYER 
// DON'T WANT TO PLAY (HE HIT ESC, SO ABORT THE GAME)
// -------------------------------------------------------------------
int CGameKernel::game_loop()
{
	int ret = CBALL_IS_DEAD;
	mtracer.add("CGameKernel::game_loop()");
	
	// Main game loop
	game_over = false; // we start just now :P
	
	speed_counter = 0;
	
	while (!game_over) 
	{
		while (speed_counter > 0)
		{
			ret = this->update_logic(); // update game logic
			speed_counter--;

			if (key[KEY_ESC] ) { game_over = true; ret = GKERNEL_USER_FINISHED_GAME; } // with ESC you end the game

			if (ret == CBALL_IS_DEAD || ret == CBALL_EXIT_LEVEL) game_over = true; // end of this game session 
		}
		
		this->update_screen(); // update screen
		frame_count++; // count fps

		if (key[KEY_F12]) // with F12 you take a screenshoot
		{
			// watermark
			textout_right(screen, font, "(C) 2004, Kronoman", SCREEN_W, SCREEN_H - text_height(font), makecol(64,64,64));
			
			save_screenshoot("kball", 0, "bmp"); // save
			
			// add confirmation message
			textout_centre(screen, font, "Screenshoot saved", SCREEN_W/2+2, SCREEN_H/2+2, makecol(255,255,255));
			clear_keybuf();
			rest(800);
			clear_keybuf();
			this->update_screen();
			speed_counter = 0;
		}
		
		if (key[KEY_ESC] || ret == GKERNEL_USER_FINISHED_GAME ) // with ESC you end the program  
		{
			// add confirmation message
			for (int yi = 0; yi < SCREEN_H; yi+=2) hline(screen,0,yi,SCREEN_W,makecol(0,0,0));
			textout_centre(screen, game_messages_font, "END GAME? Y/N", SCREEN_W/2+2, SCREEN_H/2+2, makecol(0,0,0));
			textout_centre(screen, game_messages_font, "END GAME? Y/N", SCREEN_W/2, SCREEN_H/2, makecol(0,255,255));
			clear_keybuf();
			rest(25);
			clear_keybuf();
			
			if ((readkey() >> 8) == KEY_Y)
			{
				game_over = true; 
			 	ret = GKERNEL_USER_FINISHED_GAME;
			}
			else
			{
				// game_over = false; - DEBUG, I think that this must not be here
				ret = CBALL_IS_FINE; // abort the exit 
			}
			
			this->update_screen();
			speed_counter = 0;
		}


	} // end of main game loop
	
	return ret;
}

// -------------------------------------------------------------------
// this updates 1 logic update of game ; also has code to PAUSE the game (KEY P or PAUSE)
// will return CBALL_IS_DEAD, CBALL_IS_FINE, or CBALL_EXIT_LEVEL
// -------------------------------------------------------------------
int CGameKernel::update_logic()
{
	int ret = CBALL_IS_FINE;
	
	if (key[KEY_PAUSE] || key[KEY_P]) // - pause the game -
	{
		for (int yi = 0; yi < SCREEN_H; yi+=2) hline(screen,0,yi,SCREEN_W,makecol(0,0,0));
		textout_centre(screen, game_messages_font, "- PAUSE -", SCREEN_W/2+2, SCREEN_H/2+2, makecol(0,0,0));
		textout_centre(screen, game_messages_font, "- PAUSE -", SCREEN_W/2, SCREEN_H/2, makecol(0,255,255));
		clear_keybuf();
		rest(25);
		clear_keybuf();

		readkey();

		clear_keybuf();
		rest(25);
		clear_keybuf();
		
		this->update_screen();
		speed_counter = 0;
	}
	
	game_map.update_logic();
	particle_manager.update_logic();
	
	ret = player_ball.update_logic(); // return value
	
	
	if (game_map.time_m*60+game_map.time_s <= 10) // we are running low in time!, put a warning count
	{
		static int old_time = 0;
		
		if (old_time != game_map.time_s) // only each second
		{
			old_time = game_map.time_s;
			
			char tmp_str[80];
			usprintf(tmp_str, "%d", game_map.time_s);
			
			particle_manager.add_particle(new CExplosiveTextParticle(player_ball.x+player_ball.spr->w/2, player_ball.y, (float)(rand()%100 - 50) / 100.0, (float)((rand()%300)+150) / -100.0, makecol(rand()%55+200,0,0), rand()%15+25, tmp_str, game_time_font));
		}
	}

	if (game_map.time_m*60+game_map.time_s <= 0) // time out, we are so DEAD!
	{
		ret = CBALL_IS_DEAD;
	}
	
	if (ret == CBALL_EXIT_LEVEL) // end of level :^D
	{
		if (game_map.prize_map_indispensable > 0) // we can't exit if we have prizes left.
		{
			// the user can't exit now, so drop a visual text saying it
			if (rand()%100 < 40) particle_manager.add_particle(new CTextParticle(player_ball.x+player_ball.spr->w/2, player_ball.y+player_ball.spr->h/2, (float)(rand()%300 - 150) / 100.0,(float)(rand()%300) / -100.0,makecol(rand()%55+200,0,255),rand()%25+15,string("CAN'T EXIT"), font));
			
			ret = CBALL_IS_FINE; // abort level exit
		}
	}
	
	
	
	return ret;
}

// -------------------------------------------------------------------
// This updates the screen
// -------------------------------------------------------------------
void CGameKernel::update_screen()
{
	int xc, yc; // x y of camera
		
	// camera position
	xc = (int)player_ball.x-SCREEN_W/2;
	yc = (int)player_ball.y-SCREEN_H/2;
	if (xc < 0) xc = 0;
	if (yc < 0) yc = 0;
	if (xc > TMAP_SIZE*TMAPS_W - SCREEN_W) xc = TMAP_SIZE * TMAPS_W - SCREEN_W;
	if (yc > TMAP_SIZE*TMAPS_H - SCREEN_H) yc = TMAP_SIZE * TMAPS_H - SCREEN_H;


	blit(backdropbmp, dbuffer,0,0,0,0,backdropbmp->w,backdropbmp->h); // blit background
	
	text_mode(-1); // transparent background for text

	// --- DEBUG -- REMOVE THIS MESSAGE
		textprintf(dbuffer, font, 0,SCREEN_H-text_height(font), makecol(255,255,255), "%4d fps", fps);
		textout(dbuffer,font,"WORK IN PROGRESS - NOT FINISHED DEMO",0,SCREEN_H-text_height(font)*2,makecol(255,0,0));
		textout(dbuffer,font,"Copyright (c) 2003, 2004, Kronoman",0,SCREEN_H-text_height(font)*3,makecol(128,128,255));
	// --- REMOVE UNTIL THIS LINE
	
	// ---- 3D scene (this items are draw using the Allegro's 3D software polygon routines)
	
	clear_scene(dbuffer); // clear 3D scene zbuffer
	
	// if the ball is under the floor, we must inverse the drawing queue
	//if (player_ball.z < 0)
	//{
		// actually, this is not needed, since the 3D zbuffer will take care of this 
		player_ball.draw(dbuffer,xc,yc);
		game_map.draw_map(dbuffer,xc,yc, SCREEN_W, SCREEN_H,0, false); // draw it
	//}
	//else
	//{
		//game_map.draw_map(dbuffer,xc,yc, SCREEN_W, SCREEN_H,0, false); // draw it
		//player_ball.draw(dbuffer,xc,yc);
	//}

	// draw top layer (the one with prizes =D)
	game_map.draw_map(dbuffer,xc,yc, SCREEN_W, SCREEN_H, 1, false);
	
	render_scene(); // render the 3D scene to the bitmap
		

	// ---- end of 3D scene
	
	// draw particles
	particle_manager.render(dbuffer,xc,yc);
	
	// time left message 
	textprintf_centre(dbuffer, game_time_font, dbuffer->w/2, 0, (game_map.time_m*60+game_map.time_s > 10) ? makecol(0,255,0) : makecol(255,0,0), "%02d:%02d", game_map.time_m, game_map.time_s);
	
	// score message
	textprintf(dbuffer, game_score_font, 0, 0, makecol(255,200,0), "%010d", player_ball.score);
	
	// prizes left message
	if (game_map.prize_map_indispensable > 0)
			textprintf_right(dbuffer, game_score_font, dbuffer->w, 0, makecol(0,128,255), "%04d", game_map.prize_map_indispensable);
		else
			textprintf_right(dbuffer, game_score_font, dbuffer->w, 0, makecol(0,200,255), "EXIT");
	
	// lives left message
	textprintf(dbuffer, game_score_font, 0, (int)(text_height(game_score_font)*1.05), makecol(255,220,0), "Balls:%d", player_ball.lives);

	// debug, remove this!
	textprintf(dbuffer, font,0,0,makecol(255,255,0), "x%3f,y%3f", player_ball.anglex,player_ball.angley);

	blit(dbuffer, screen,0,0,0,0,dbuffer->w,dbuffer->h); // dump the render to screen


	
	yield_timeslice(); // play nice with multitask
}

// -------------------------------------------------------------------
// Helper function, takes a bitmap, and returns a new allocated bitmap
// with the 1st bitmap tiled in the size of the screen
// this serves for the background imagen
// -------------------------------------------------------------------
BITMAP *CGameKernel::tile_bmp_to_screen_size(BITMAP *bmp)
{
	BITMAP *b = NULL;
	if (bmp == NULL) raise_error("CGameKernel::tile_bmp_to_screen_size()\nERROR: bitmap null; can't tile it!\n");

	b = create_bitmap(SCREEN_W, SCREEN_H);
	if (b == NULL) raise_error("CGameKernel::tile_bmp_to_screen_size()\nERROR: unable to create bitmap!\n");
	
	for (int y = 0; y < SCREEN_H+bmp->h; y+=bmp->h)
		for (int x = 0; x < SCREEN_W+bmp->w; x+=bmp->w)
			blit(bmp, b,0,0,x,y,bmp->w,bmp->h);
	
	return b;
}

// -----------------------------------------------------------------------
// Loads a level from a file, and sets the game ready to play on that level
// 
// *MUST* BE CALLED BEFORE STARTING THE GAME LOOP!
// WILL LOAD THE MAP, BACKGROUND, etc
// -----------------------------------------------------------------------
void CGameKernel::load_level_file(const char *file) 
{
	mtracer.add("CGameKernel::load_level_file(\"%s\")", file);
	
	if (game_map.load_map_from_file(file)) raise_error("CGameKernel::load_level_file()\nERROR!\nCan't load level '%s'!\n", file);
	
	usprintf(current_level_file_name, "%s", file); // save current file name
	
	// set the player position and properties
	player_ball.x = (game_map.sxp * TMAPS_W)+(TMAPS_W - BALL_RADIUS*2)/2;
	player_ball.y = (game_map.syp * TMAPS_H)+(TMAPS_H - BALL_RADIUS*2)/2;
	player_ball.z = 1.0;
	player_ball.dx = player_ball.dy = player_ball.dz = 0.0;
	player_ball.ctmap = &game_map;
	player_ball.particle_manager = &particle_manager;
	
	particle_manager.nuke_particle_list(); // CLEAN THE PARTICLES FOR THE NEW LEVEL :P
	
	// DEBUG -- BACKGROUND! - DECIDIR QUE CORNO HAGO CON ESTO
	// Load the background from info stored on the level map
	if (backdropbmp != NULL)
	{
		destroy_bitmap(backdropbmp);
		backdropbmp = NULL;
	}
	
	
	backdropbmp = tile_bmp_to_screen_size( background_loader.get_background("backgr.dat", game_map.background_index) );
	
	mtracer.add("\tLoaded background of level OK");
}


// -----------------------------------------------------------------------
// This will start a single level game
// Will return CBALL_IS_DEAD(totally DEAD, full game over, 0 lives)
// CBALL_EXIT_LEVEL (only useful for campaign), or GKERNEL_USER_FINISHED_GAME if player cancelled game
// -----------------------------------------------------------------------
int CGameKernel::play_a_single_level(char *level_filename)
{
	int ret;
	bool do_the_loop = true; // loop, for loading level many times, etc
	int old_score = 0; // this is for restoring the score when the player loses a ball
	
	mtracer.add("\nCGameKernel::play_a_single_level('%s');\n", level_filename);

	this->init(); // init game kernel
	
	old_score = player_ball.score;
	
	while (do_the_loop)
	{
		mtracer.add("\tLoading and playing level.");
		
		this->load_level_file(level_filename);
		
		ret = this->game_loop();
		
		mtracer.add("\t--> this->game_loop = %4d\n", ret);
		
		// according to ret value, show the player, if we won, we lose, or if we have lives to keep playing, etc
		// the { } in the case <x>: are there to force the scope, so don't remove them (I know what I'm doing...)
		switch (ret)
		{
			case CBALL_EXIT_LEVEL:
			{
				// the player won the level, OK, so we exit
				// this value will be used by campaign function, if available
				
				this->update_screen();
				for (int yi = 0; yi < SCREEN_H; yi+=2) hline(screen,0,yi,SCREEN_W,makecol(0,0,0));
				
				textout_centre(screen, game_messages_font, "YOU WON THE LEVEL!", SCREEN_W/2+2, SCREEN_H/2+2, makecol(0,0,0));
				textout_centre(screen, game_messages_font, "YOU WON THE LEVEL!", SCREEN_W/2, SCREEN_H/2, makecol(0,255,255));
				
				// GIVE BONUS
				int bonus = 0;
				text_mode(makecol(0,0,64));
				
				clear_keybuf();
				rest(550);
				clear_keybuf();
				
				rectfill(screen,0,SCREEN_H/2+text_height(game_messages_font), SCREEN_W, SCREEN_H, makecol(0,0,64));
				
				
				textout_centre(screen, game_messages_font, "TIME BONUS", SCREEN_W/2, SCREEN_H/2+text_height(game_messages_font), makecol(0,255,255));
	
				for (int i = 0; i < game_map.time_m*60+game_map.time_s; i++)
				{
					textprintf_centre(screen, game_messages_font, SCREEN_W/2, SCREEN_H/2+text_height(game_messages_font)*2, makecol(0,255,255),"   %04d x 100 = %07d   ", i, bonus);
					bonus += 100; // 100 bonus points for each second
					
					// DEBUG -- falta SONIDO!
					
					if (keypressed()) break;
					rest(10);
				}
				
				// give real bonus
				bonus = (game_map.time_m*60+game_map.time_s)*100;
				
				player_ball.score += bonus;
				
				textprintf_centre(screen, game_messages_font, SCREEN_W/2, SCREEN_H/2+text_height(game_messages_font)*2, makecol(0,255,255),"   %04d x 100 = %07d   ", game_map.time_m*60+game_map.time_s, bonus);
				textprintf_centre(screen, game_messages_font, SCREEN_W/2, SCREEN_H/2+text_height(game_messages_font)*3, makecol(0,255,255),"Score = %010d", player_ball.score);
				
				text_mode(-1);
				
				rest(700);
				textout_centre(screen, font, "-- Hit any key to continue --", SCREEN_W/2, SCREEN_H-text_height(font), makecol(255,255,255));
				
				clear_keybuf();
				readkey();
				clear_keybuf();
				
				
				do_the_loop = false;
			}
			break; // CBALL_EXIT_LEVEL
			
			
			case CBALL_IS_DEAD:
			{
				this->update_screen();
				
				for (int yi = 0; yi < SCREEN_H; yi+=2) hline(screen,0,yi,SCREEN_W,makecol(0,0,0));
				
				if (game_map.time_m*60+game_map.time_s > 0)
				{
					textout_centre(screen, game_messages_font, "YOU LOST THE BALL!", SCREEN_W/2+2, SCREEN_H/2+2, makecol(0,0,0));
					textout_centre(screen, game_messages_font, "YOU LOST THE BALL!", SCREEN_W/2, SCREEN_H/2, makecol(0,255,255));
				}
				else
				{
					textout_centre(screen, game_messages_font, "TIME UP!", SCREEN_W/2+2, SCREEN_H/2+2, makecol(0,0,0));
					textout_centre(screen, game_messages_font, "TIME UP!", SCREEN_W/2, SCREEN_H/2, makecol(0,255,255));
				}
				
				rest(700);
				textout_centre(screen, font, "-- Hit any key to continue --", SCREEN_W/2, SCREEN_H-text_height(font), makecol(255,255,255));
				
				clear_keybuf();
				readkey();
				clear_keybuf();
				
				// the player died, rest one live
				player_ball.lives--;
				player_ball.anglex = player_ball.angley = 0.0; // reset ball rotations
				if (player_ball.lives < 0) // the player is dead. 0 lives. game over. :(
				{
					
					BITMAP *bmp_game_over = (BITMAP *)main_data_file.get_resource_dat("GAME_OVER_BMP");
	
					blit(bmp_game_over, screen, 0,0,0,0,bmp_game_over->w,bmp_game_over->h);
					rest(700);
					textout_centre(screen, font, "-- Hit any key to continue --", SCREEN_W/2, SCREEN_H-text_height(font), makecol(128,0,0));				
					clear_keybuf();
					readkey();
					clear_keybuf();
					
					do_the_loop = false;
				}
				else
				{
					// perdio una vida, pero le queda, jugar de nuevo el nivel
					do_the_loop = true; // load the next level, damn
					player_ball.score = old_score; // restore score
				}
			}	
			break; // end of CBALL_IS_DEAD
			
			case GKERNEL_USER_FINISHED_GAME:
				do_the_loop = false; // the player hit <ESC>, end of game :(
			break; // GKERNEL_USER_FINISHED_GAME
			
			default: // it should NEVER come here
				raise_error("ERROR at CGameKernel::play_a_single_level\nGuru meditation:\ngame_loop returned a invalid value (ret = %d)\n", ret);
			break;
		}
		
	}

	this->shutdown(); // shutdown this beauty...
	
return ret;
}


// -----------------------------------------------------------------------
// This will start a full campaign game
// -----------------------------------------------------------------------
void CGameKernel::play_a_full_campaign(char *level_filename)
{
	char tmp_str[4096]; // buffer to handle level filenames
	int current_level = 1; // current level number
	bool finished = false;
	int ret = 0;
	
	mtracer.add("\nCGameKernel::play_a_full_campaign('%s')\n", level_filename);
	
	while (!finished)
	{
		// load and play level	
		usprintf(tmp_str,"%s#%d_MAP", level_filename, current_level);
		fix_filename_slashes(tmp_str);

		// check if the level exits before trying to play it -- DEBUG -- slow code here! do some better
		PACKFILE *fp_test = pack_fopen(tmp_str, F_READ);
		if (fp_test == NULL)
		{
			// EL NIVEL NO EXISTE, LISTO, GANO
			
			// the player won the campaign
			
			// DEBUG -- falta mensaje
			
			return; // chau... finalizo!
		}
		
		pack_fclose(fp_test);
		
		ret = this->play_a_single_level(tmp_str);

		mtracer.add("\t--> this->play_a_single_level('%s') = %4d\n",tmp_str, ret);
		
		if (ret == CBALL_EXIT_LEVEL) // exit level!
		{
			current_level++;
			finished = false;
		}
		else
		{
			finished = true; // any other value means GAME OVER, sorry d00d, insert coin :P	
		}
		
	}
}

